import os

from elasticsearch import ElasticsearchException, ConnectionTimeout, Elasticsearch, NotFoundError, TransportError
from log_config import logger

from service_managment import DockerComposeServiceManager, ServiceId

import conduit_config
import time
import json
from datetime import datetime, timezone


class BackupService:

    def __init__(self, _es_client: Elasticsearch, _service_management: DockerComposeServiceManager):
        self._service_management = _service_management
        self._es_client = _es_client

    def create_repository_and_policy(self, json_body_request):
        self.__create_es_repository_and_policy(json_body_request)

    def get_es_repository(self, repo_name: str = None):
        """
        Retrieve a dictionary with elasticsearch repositories. If repo name is specified, will retrieve only one
        repository
        :rtype:
        """
        if repo_name is None:
            logger.info("Retrieving details about all repositories")
        else:
            logger.info("Retrieving details about {} repository".format(repo_name))

        try:
            es_repo = self._es_client.snapshot.get_repository(repository=repo_name)
            if repo_name is not None:
                logger.info("Retrieved repository {} successfully".format(repo_name))

                repo = es_repo[repo_name]
                logger.info("Repository data: {}".format(str(es_repo)))

                self.verify_repository(repo_name)
                logger.info("Verified successfully repository {}".format(repo_name))

            else:
                logger.info("Retrieved all repositories successfully")
            return es_repo

        except NotFoundError:
            return None

    def create_snapshot(self, repo_name, backup_name, snapshot_settings):
        self.__create_es_snapshot(repo_name, backup_name, snapshot_settings)

    def create_lifecycle(self, policy_id, body):
        logger.info("Creating lifecycle {}".format(policy_id))
        create_slm_response = self._es_client.slm.put_lifecycle(policy_id, body=body)
        if not create_slm_response or create_slm_response['acknowledged'] is not True:
            logger.warning("Could not delete snapshot policy. %s." % str(create_slm_response))
            return False
        logger.info("Lifecycle created successfully")
        return True

    def get_lifecycle(self, policy_id=None):
        if policy_id is None:
            logger.info("Retrieving information about all snapshot lifecycle policies")
        else:
            logger.info("Retrieving information about snapshot lifecycle policy {}".format(policy_id))
        try:
            res = self._es_client.slm.get_lifecycle(policy_id=policy_id)
            if policy_id is None:
                logger.info("Successfully retrieved information about all snapshot lifecycle policies")
            else:
                logger.info("Successfully retrieved information about snapshot lifecycle policy {}".format(policy_id))
                try:
                    d = datetime.now(timezone.utc).astimezone()
                    utc_offset = d.utcoffset()
                    res[policy_id]["timezone_offset"] = utc_offset.seconds/3600
                except Exception as e:
                    logger.warning("Could not obtain timezone offset. {}".format(str(e)))
            return res
        except NotFoundError:
            logger.warning("Snapshot Lifecycle Policy {} not found".format(policy_id))
            pass
        return None

    def verify_repository(self, repo_name):
        logger.info("Verifying repository {}".format(repo_name))
        try:
            self._es_client.snapshot.verify_repository(repository=repo_name)
        except TransportError as te:
            if te.error == 'repository_verification_exception':
                logger.exception("Invalid repository details. "
                                 "Please configure your repository details. {}".format(str(te.info)))
                raise RepositoryVerificationError(te.status_code, te)
            raise te

    def execute_lifecycle(self, policy_id, repository_name):
        """
         Immediately creates a snapshot according to the lifecycle policy, without waiting for the scheduled time.
        """
        logger.info("Executing backup snapshot for {} lifecycle".format(policy_id))
        self.get_es_repository(repository_name)
        lifecycle_policy_exists = self.get_lifecycle(policy_id)
        if lifecycle_policy_exists is None:
            logger.warning("Could not execute lifecycle snapshot on demand. {} lifecycle not found".format(policy_id))
            return None
        res = self._es_client.slm.execute_lifecycle(policy_id)
        self.get_snapshots(res["snapshot_name"])
        logger.info("Successfully executed snapshot for {} policy".format(policy_id))
        return res

    def get_snapshots(self, snapshot_name=None):
        if snapshot_name is None:
            logger.info("Retrieving all snapshots")
            _snapshot_name = "_all"
        else:
            _snapshot_name = snapshot_name
            logger.info("Retrieve {} snapshot".format(snapshot_name))
        try:
            snapshots_results = self._es_client.snapshot.get(repository="conduit_snapshot_repository",
                                                             snapshot=_snapshot_name,
                                                             pretty=True,
                                                             request_timeout=600,
                                                             ignore_unavailable=True)
            results = snapshots_results["snapshots"]
            results.sort(reverse=True, key=self.sort_snapshots_by_end_time)
            if _snapshot_name == '_all':
                logger.info("Found {} snapshots".format(len(results)))
            else:
                logger.info("Found {} snapshot".format(_snapshot_name))
            return results
        except ElasticsearchException as ee:
            logger.exception("Failed to retrieve ES data snapshot {}".format(_snapshot_name), exc_info=ee)
            if str(ee).find("concurrently modifying the contents") > 0 or str(ee).find("concurrent modification") > 0:
                raise RepositoryConcurrencyError(ee.status_code, ee)
            else:
                raise ee

    def __wait_for_service_to_start(self, service_name):
        service_id = ServiceId(service_name)
        status = self._service_management.service_status(service_id)
        start_time = datetime.now()
        while status.is_healthy is False:
            time.sleep(5)
            status = self._service_management.service_status(service_id)
            # After ten minutes, we should throw an error
            elapsed_in_seconds = (datetime.now() - start_time).total_seconds()
            if elapsed_in_seconds > 600:
                raise ConnectionTimeout(
                    500, "Could not start {} container".format(service_name))
        logger.info("Service {} is up and running".format(service_name))

    def validate_repository_request(self, json_body_request):
        logger.info("Creating es repository. Validating request object")
        if "snapshotRepository" not in json_body_request.keys():
            return "'snapshotRepository' field is required"
        repository = json_body_request["snapshotRepository"]

        if "settings" not in repository.keys():
            return "'settings' field is required"

        if "type" not in repository.keys():
            return "'type' field is required"
        else:
            if repository["type"] != "azure" and repository["type"] != "fs" and repository["type"] != "gcs" and repository["type"] != "s3":
                return "'type' field not valid. It must have one of following values [azure, hdfs, fs, gcs, s3]"

        repository_type = repository["type"]
        repository_settings_request = repository["settings"]

        if repository_type == 'azure':
            if "container" not in repository_settings_request.keys():
                return "'container' field is required for azure repository"
            if "storageAccount" not in repository_settings_request.keys():
                return "'storageAccount' field is required for azure repository"
            if "base_path" not in repository_settings_request.keys():
                return "'base_path' field is required for azure repository"
            if "storageKey" not in repository_settings_request.keys():
                return "'storageKey' field is required for azure repository"

        if repository_type == 'gcs':
            if "bucket" not in repository_settings_request.keys():
                return "'bucket' field is required for gcs repository"
            if "base_path" not in repository_settings_request.keys():
                return "'base_path' field is required for gcs repository"
            if "credentials" not in repository.keys():
                return "'credentials' field is required for gcs repository"
            gcs_credentials = repository["credentials"]
            if "type" not in gcs_credentials.keys():
                return "'type' field is required for gcs repository credentials"
            if "project_id" not in gcs_credentials.keys():
                return "'project_id' field is required for gcs repository credentials"
            if "private_key_id" not in gcs_credentials.keys():
                return "'private_key_id' field is required for gcs repository credentials"
            if "private_key" not in gcs_credentials.keys():
                return "'private_key' field is required for gcs repository credentials"
            if "client_email" not in gcs_credentials.keys():
                return "'client_email' field is required for gcs repository credentials"
            if "client_id" not in gcs_credentials.keys():
                return "'client_id' field is required for gcs repository credentials"

        if repository_type == 's3':
            if "bucket" not in repository_settings_request.keys():
                return "'bucket' field is required for s3 repository"
            if "accessKey" not in repository_settings_request.keys():
                return "'accessKey' field is required for s3 repository"
            if "region" not in repository_settings_request.keys():
                return "'region' field is required for s3 repository"
            if "base_path" not in repository_settings_request.keys():
                return "'base_path' field is required for s3 repository"
            if "secretKey" not in repository_settings_request.keys():
                return "'secretKey' field is required for s3 repository"

        logger.info("Repository request object validated successfully")

        return None

    def __create_es_repository_and_policy(self, json_body_request):
        repository = json_body_request["snapshotRepository"]
        repository_settings_request = repository["settings"]
        repository_type = repository["type"]
        repository_name = conduit_config.get_elasticsearch_repository()

        repository_settings = {"type": repository_type}

        if repository_type == 'fs':
            path = conduit_config.get_container_snapshots_location()
            logger.info("Creating local es repository {} in {} folder of local system".format(repository_name, path))
            repository_settings["settings"] = {
                "host_location": conduit_config.get_system_snapshots_location(),
                "location": path
            }

        if repository_type == 'azure':
            container = repository_settings_request["container"]
            repository_account = repository_settings_request["storageAccount"]
            path = repository_settings_request["base_path"]
            logger.info("Creating azure es repository {}, in {} folder, of {} container with {} storage account".format(repository_name, path, container, repository_account))
            repository_settings['settings'] = {
                "base_path": path,
                "container": container,
                "repository_account": repository_account
            }
            repository_key = repository_settings_request["storageKey"]
            if repository_account is not None and repository_key is not None:
                self.__create_azure_keystore_entries(repository_account, repository_key)
            else:
                logger.exception("Cannot create backup repository. Azure storage account or access key missing")
                raise ElasticsearchException("Cannot create backup repository. Azure storage account or secretKey missing")

        if repository_type == 's3':
            bucket = repository_settings_request["bucket"]
            repository_account = repository_settings_request["accessKey"]
            region = repository_settings_request["region"]
            path = repository_settings_request["base_path"]
            logger.info("Creating aws s3 es repository {}, in {} folder of {} bucket of {} region with {} access key"
                        .format(repository_name, path, bucket, region, repository_account))

            repository_settings["settings"] = {
                "base_path": path,
                "bucket": bucket,
                "region": region,
                "repository_account": repository_account
            }

            repository_key = repository_settings_request["secretKey"]
            if repository_account is not None and repository_key is not None:
                self.__create_s3_keystore_entries(repository_account, repository_key)
            else:
                logger.exception("Cannot create backup repository. AWS S3 accessKey or secretKey missing")
                raise ElasticsearchException("Cannot create backup repository. AWS S3 accessKey or secretKey missing")

        if repository_type == 'gcs':
            bucket = repository_settings_request["bucket"]
            location = repository_settings_request["base_path"]
            gcs_credentials = repository["credentials"]
            repository_account = gcs_credentials["client_email"]
            logger.info("Creating GCS es repository {} on {} bucket in {} folder with {} account"
                        .format(repository_name, bucket, location, repository_account))

            repository_settings["settings"] = {
                "base_path": location,
                "bucket": bucket,
                "repository_account": repository_account
            }

            if gcs_credentials is not None and repository_account is not None:
                self.__create_gcs_keystore_entries(gcs_credentials)
            else:
                logger.exception("Cannot create backup repository. AWS S3 accessKey or secretKey missing")
                raise ElasticsearchException("Cannot create backup repository. AWS S3 accessKey or secretKey missing")

        repository_settings["compress"] = True


        try:
            create_repo_response = self._es_client.snapshot.create_repository(
                repository=repository_name, body=repository_settings, verify=True)

            logger.debug("Create repository response: " + str(create_repo_response))
            if not create_repo_response or create_repo_response['acknowledged'] is not True:
                logger.exception("Could not create repository. {}".format(create_repo_response))
                raise ElasticsearchException("Could not create repository. {}".format(str(create_repo_response)))
            logger.info("Repository {} created successfully".format(repository_name))
        except TransportError as te:
            if te.error == 'repository_verification_exception':
                logger.exception("Invalid repository details. "
                         "Please configure your repository details. {}".format(str(te.info)))
                raise RepositoryVerificationError(te.status_code, te)
            else:
                logger.exception("Failed to create ES repository {}. {} ".format(repository_name, str(te)))
                raise te
        except ElasticsearchException as ee:
            logger.exception("Failed to create ES repository {}. {} ".format(repository_name, str(ee)))
            raise ee

        if 'snapshotPolicyManagement' in json_body_request.keys():
            snapshot_policy = json_body_request["snapshotPolicyManagement"]
            snapshot_lifecycle_policy = {
                "schedule": snapshot_policy["schedule"],
                "name": "<conduit-snap-{now}>",
                "repository": repository_name,
                "config": {
                    "indices": conduit_config.get_system_backup_indexes(),
                    "ignore_unavailable": True,
                    "include_global_state": False
                },
                "retention": {
                    "expire_after": "90d",
                    "min_count": 5,
                    "max_count": 50
                }
            }
            logger.info("Creating snapshot lifecycle management for {} repository. Policy used: {}".format(repository_name, str(snapshot_lifecycle_policy)))
            self.create_lifecycle("conduit-snapshot", snapshot_lifecycle_policy)
            logger.info("Created successfully snapshot lifecycle management for {} repository".format(repository_name))

        else:
            logger.info("Removing existing snapshot lifecycle management for {} repository".format(repository_name))
            try:
                self._es_client.slm.delete_lifecycle("conduit-snapshot")
            except NotFoundError:
                pass

    def __create_es_snapshot(self, repo_name, backup_name, snapshot_settings):
        try:
            logger.info("Creating es snapshot for {} repository".format(repo_name))
            es_repo = self.get_es_repository(repo_name)
            if not es_repo:
                logger.exception("{} repository not found , create one first".format(repo_name))
                raise ElasticsearchException("{} repository not found , create one first".format(repo_name))
            create_response = self._es_client.snapshot.create(repository=repo_name,
                                                              request_timeout=600,
                                                              snapshot=backup_name, body=snapshot_settings,
                                                              wait_for_completion=True)
            if not create_response or create_response['snapshot']['state'] != "SUCCESS":
                raise ElasticsearchException("Snapshot was not successfully created.")

            logger.info("Snapshot created successfully {}".format(backup_name))
            return create_response

        except ElasticsearchException as ee:
            logger.exception("Failed to create ES data snapshot {}".format(backup_name), exc_info=ee)
            if str(ee).find("concurrently modifying the contents") > 0 or str(ee).find("concurrent modification") > 0:
                raise RepositoryConcurrencyError(ee.status_code, ee)
            else:
                raise ee

    def __create_azure_keystore_entries(self, account: str, key: str):

        account_cmd = '{} azure.client.default.account {}'.format(conduit_config.get_es_keystore_importer_path(),
                                                                  account)
        key_cmd = '{} azure.client.default.key {}'.format(conduit_config.get_es_keystore_importer_path(), key)
        self.__insert_es_keystore_values(account_cmd, key_cmd)

    def __create_s3_keystore_entries(self, account: str, key: str):
        account_cmd = '{} s3.client.default.access_key {}'.format(conduit_config.get_es_keystore_importer_path(),
                                                                  account)
        key_cmd = '{} s3.client.default.secret_key {}'.format(conduit_config.get_es_keystore_importer_path(), key)
        self.__insert_es_keystore_values(account_cmd, key_cmd)

    def __create_gcs_keystore_entries(self, credentials: str):
        logger.info("Creating GCS elasticsearch-keystore entry")
        credentials_file_name = "gcs.json"
        credentials_file_container_path = os.path.join(conduit_config.get_container_backup_location(),
                                                       credentials_file_name)
        path = os.path.join(conduit_config.get_system_backup_location(), credentials_file_name)
        logger.info("Creating GCS elasticsearch-keystore credentials file in {} on host "
                    "and in {} on container".format(path, credentials_file_container_path))
        f = open(path, "w")
        json.dump(credentials, f)
        f.close()
        logger.info("GCS credentials file for es-keystore created successfully. Will insert it into es-keystore")

        keystore_add_file_command = "elasticsearch-keystore add-file -f gcs.client.default.credentials_file {}" \
            .format(credentials_file_container_path)
        self._service_management.exec_cmd(ServiceId("conduit-storage"), keystore_add_file_command)
        logger.info("Inserted successfully gcs credentials file to es-keystore")

        logger.info("Reload elasticsearch secure settings to load sensitive data ( account & password )")
        self._es_client.nodes.reload_secure_settings()
        logger.info("Reloaded successfully elasticsearch secure settings")

    def __insert_es_keystore_values(self, account_cmd: str, key_cmd: str):
        logger.info("Try to insert elasticsearch repository sensitive values in elasticsearch-keystore")
        service_id = ServiceId("conduit-storage")
        self._service_management.exec_cmd(service_id, account_cmd)
        self._service_management.exec_cmd(service_id, key_cmd)
        logger.info("Reload elasticsearch secure settings to load sensitive data ( account & password )")
        self._es_client.nodes.reload_secure_settings()
        logger.info("Reloaded successfully elasticsearch secure settings")


    def sort_snapshots_by_end_time(self, e):
        return e['end_time_in_millis']


class RepositoryVerificationError(Exception):

    def __init__(self, error_code, *args, **kwargs):
        super(Exception, *args, **kwargs)
        self.status_code = error_code

class RepositoryConcurrencyError(Exception):

    def __init__(self, error_code, *args, **kwargs):
        super(Exception, *args, **kwargs)
        self.status_code = error_code
