#!/usr/bin/env bash

SUDO_ACCESS=$(id -u)
if [ "$SUDO_ACCESS" -ne 0 ]; then
  echo "Root permission required to execute Conduit restore"
  exit 2
fi

REPOSITORY_NAME="conduit_snapshot_repository"
DISTRIBUTION=$(uname -s)
LINUX_DISTRIBUTION="Linux"

if [[ "$DISTRIBUTION" == "$LINUX_DISTRIBUTION" ]] ; then
  if ! command -v ansible &> /dev/null
  then
    echo "Installing ansible "
    apt-get install -y software-properties-common whois jq at \
                       python3 python3-pip python3-apt python3-setuptools
    python3 -m pip install --upgrade pip
    python3 -m pip install ansible==2.9.13

  fi
fi

execute_restore() {
    repo_name="${1:-conduit_snapshot_repository}"

    echo "Repository created. We'll soon display available snapshots names."

    python3 get_snapshots.py "$repo_name"
    python_response=$?
    if [ $python_response -ne 0 ]; then
      exit 2
    fi
    ansible-playbook ./conduit-restore.yml -e repository_name="$repo_name"
    ansible_response=$?
    if [ $ansible_response -ne 0 ]; then
        echo "Restore could not be done. Please contact administrator"
        exit 2
    fi
    CONDUIT_RESTORE_RESPONSE=$?
    if [ $CONDUIT_RESTORE_RESPONSE -eq 0 ]; then
      echo "Restore was successful"
      exit 0
    else
      echo "Restore could not be done. Please contact administrator"
      exit 2
    fi
}

create_repository() {
  read_only="${1:-false}"
  repo_name="${2:-conduit_snapshot_repository}"

  echo "Enter repository type (fs / azure / s3 / gcs)"
  read -r repository_type_input

  repository_type=$(echo "$repository_type_input" | tr '[:upper:]' '[:lower:]')
  case $repository_type in

    s3)
      ansible-playbook ./conduit-restore-s3-repository.yml -e "read_only=$read_only" -e "repository_name=$repo_name"
      ;;

    azure)
      ansible-playbook ./conduit-restore-azure-repository.yml -e "read_only=$read_only" -e "repository_name=$repo_name"
      ;;

    gcs)
      ansible-playbook ./conduit-restore-gcs-repository.yml -e "read_only=$read_only" -e "repository_name=$repo_name"
      ;;

    fs)
      ansible-playbook ./conduit-restore-fs-repository.yml -e "read_only=$read_only" -e "repository_name=$repo_name"
      ;;

    *)
      echo "Unknown value. Please start the script again and enter one of allowed repositories (fs / azure / s3)"
      exit 2
  esac
  ansible_response=$?
  if [ $ansible_response -ne 0 ]; then
      echo "Failed create repository in elasticsearch"
      exit 2
  fi

}

verify_existing_repository() {
  repo_name="${1:-"conduit_snapshot_repository"}"
  VERIFY_REPOSITORY_CODE=$(curl -X POST --write-out '%{http_code}' --silent --output /dev/null "http://localhost:9200/_snapshot/${repo_name}/_verify")
  if [ "$VERIFY_REPOSITORY_CODE" -eq 503 ]; then
    echo "Elasticsearch container is not up & running, please start it and run again"
    exit 2
  fi

  if [ "$VERIFY_REPOSITORY_CODE" -eq 405 ] ||
     [ "$VERIFY_REPOSITORY_CODE" -eq 301 ] ||
     [ "$VERIFY_REPOSITORY_CODE" -eq 000 ]; then
    echo "Elasticsearch could not verify repository. Please contact administrator"
    exit 2
  fi

  if [ "$VERIFY_REPOSITORY_CODE" -eq 200 ]; then
    echo 0
  fi

  if [ "$VERIFY_REPOSITORY_CODE" -eq 404 ] ||
     [ "$VERIFY_REPOSITORY_CODE" -eq 500 ]; then
    echo 1
  fi

}

create_missing_repository() {
    create_repository true conduit_restore_repository
    CONDUIT_CREATE_BACKUP_REPOSITORY=$?

    if [ $CONDUIT_CREATE_BACKUP_REPOSITORY -eq 0 ]; then
      execute_restore conduit_restore_repository
    else
      echo Repository could not be created. Please contact administrator
      exit 2
    fi
}
VERIFY_REPOSITORY_CODE=$(verify_existing_repository "conduit_snapshot_repository")
if [ "$VERIFY_REPOSITORY_CODE" -eq 0 ]; then
  echo "Do you want to restore from current configured repository or to create a new one ?"
  echo "'YES' to use current configured repository or 'NO' to use another repository"
  read -r create_repository_response_input
  create_repository_response=$(echo "$create_repository_response_input" | tr '[:upper:]' '[:lower:]')
  case $create_repository_response in

  yes)
    if [ "$VERIFY_REPOSITORY_CODE" -eq 0 ]; then
      execute_restore conduit_snapshot_repository
    fi
    ;;
  no)
    echo "Please enter details to create a repository"
    create_missing_repository
    ;;
  *)
    echo "Unknown value. Please start the script again and enter one of allowed responses (yes / no)"
    exit 2
esac
else
  echo "Repository missing, please enter details to recreate repository"
  create_missing_repository
fi

