import json
import dateutil.parser
import requests

import socket

from bottle import Bottle
from bottle import HTTPResponse
from bottle import request, response
from compose.project import NoSuchService
from docker import errors
import conduit_config
from requests import exceptions

from elasticsearch import Elasticsearch, ElasticsearchException, ConnectionTimeout, ConnectionError, NotFoundError, \
    TransportError

from common import UpgradeContext, DeploymentStatus
from backup_service import BackupService, RepositoryVerificationError, RepositoryConcurrencyError
from infrastructure_api import DeploymentApiClient
from log_config import logger
from service_managment import DockerComposeServiceManager, ServiceManager, ServiceId, ServiceStatus
from upgrade_service import UpgradeService, UpgradeError, ConcurrentUpgradeAttempt
from certificate_manager import CertificateInstaller
from certificate_manager import CertificateInstallationError
app = Bottle()

server_ip = conduit_config.get("service.server.ip")
server_port = int(conduit_config.get("service.server.port"))

deployment_api_url = conduit_config.get("system.deployment_api.url")
deploy_service_api = DeploymentApiClient(deployment_api_uri=deployment_api_url,
                                         auth_token=conduit_config.get_deployment_token())

es_client_ = Elasticsearch(hosts=conduit_config.get_elasticsearch_host(), timeout=60)

_service_management = DockerComposeServiceManager(base_dir=conduit_config.get_docker_compose_dir(),
                                                  filenames=conduit_config.get_docker_compose_files())
_backup_service = BackupService(es_client_, _service_management)

@app.route('/')
def index():
    return ' { "status": "ok", "app_version": "1.0" } '


@app.route('/deployment-status', method='GET')
def deployment_status():
    resp = deploy_service_api.get_deployment(conduit_config.get_conduit_license_token())
    if resp:
        data = {
            "status": resp['status'],
            "currentVersion": resp['currentVersion']
        }
        response.headers['Content-Type'] = 'application/json'
        return json.dumps(data)
    else:
        return HTTPResponse(status=404)


@app.route('/versions', method='GET')
def list_available_version():
    logger.debug("getting list of versions")
    try:
        versions = []
        try:
            versions = deploy_service_api.list_versions()
        except exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return HTTPResponse(status=404, headers={'Content-Type': 'application/json'},
                                    body={"error": "No available versions"})
            else:
                logger.exception("Failed to load versions", exc_info=e)
                return HTTPResponse(status=e.response.status_code, headers={'Content-Type': 'application/json'},
                                    body={"error": "Failed to load versions"})

        current_version = None
        try:
            current_version = deploy_service_api.get_current_version(conduit_config.get_conduit_license_token())
        except exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return HTTPResponse(status=404, headers={'Content-Type': 'application/json'},
                                    body={"error": "No current version registered"})
            else:
                logger.exception("Failed to get current version", exc_info=e)
                return HTTPResponse(status=e.response.status_code, headers={'Content-Type': 'application/json'},
                                    body={"error": "Failed to load current version"})

        current_version_def = [v for v in versions if v["version"] == current_version]
        available_versions = []
        if len(current_version_def) == 1 and current_version_def[0]['releaseDate'] and current_version_def[0]['releaseDate'] != "":
            current_release_date = dateutil.parser.parse(current_version_def[0]['releaseDate'])
            available_versions = [v for v in versions if v['releaseDate'] and dateutil.parser.parse(v['releaseDate']) > current_release_date]

        response.headers['Content-Type'] = 'application/json'
        return json.dumps(available_versions)
    except BaseException as re:
        logger.exception(" deployment-service returned error ", exc_info=re)
        return HTTPResponse(status=500, headers={'Content-Type': 'application/json'},
                            body={"error": "Couldn't load available versions"})


@app.route('/upgrade/<target_version>', method='POST')
def run_update(target_version):
    current_version = deploy_service_api.get_current_version(conduit_config.get_conduit_license_token())
    if not current_version:
        logger.warning("No current version in deployment database")
        return HTTPResponse(status=404, headers={'Content-Type': 'application/json'},
                            body={"error": "No version found"})
    if current_version == target_version:
        logger.debug("No migration required, version %s " % current_version)
        return HTTPResponse(status=400, headers={'Content-Type': 'application/json'},
                            body={"error": "Nothing to upgrade, current version is same"})

    target_version_def = deploy_service_api.get_version_def(target_version)
    if not target_version_def:
        return HTTPResponse(status=400, headers={'Content-Type': 'application/json'},
                            body={"error": "version %s not found" % target_version})
    current_status = deploy_service_api.get_deployment_status(conduit_config.get_conduit_license_token())
    if current_status and current_status.is_in_progress():
        return HTTPResponse(status=409, headers={'Content-Type': 'application/json'},
                            body={"error": "upgrade is in progress"})
    payload = request.json or dict()
    dry_run = _to_bool(payload.get("dry_run", request.query.dry_run), False)
    try:
        update_service = UpgradeService(deploy_service_api)
        update_service.execute_update(conduit_config.get_conduit_license_token(), current_version,
                                      target_version, dry_run)
    except UpgradeError as ue:
        logger.exception("Failed to update ", exc_info=ue.cause)
        if not dry_run:
            # set failure only iff it's not a dry run
            set_upgrade_failure_reason(ue.update_context, ue.cause)
            return HTTPResponse(status=500,
                                body={"error": "unexpected error during upgrade. Please contact administrator"})
        else:
            return HTTPResponse(status=500, body={"error": "release verification error"})
    except ConcurrentUpgradeAttempt:
        return HTTPResponse(status=409, headers={'Content-Type': 'application/json'},
                            body={"error": "upgrade is in progress"})
    except ValueError as ve:
        logger.exception("Illegal argument ", exc_info=ve)
        return HTTPResponse(status=400, headers={'Content-Type': 'application/json'},
                            body={"error": str(ve)})
    except BaseException as be:
        logger.exception("Failed to update ", exc_info=be)
        return HTTPResponse(status=500, headers={'Content-Type': 'application/json'},
                            body={"error": str(be)})
    if dry_run:
        return HTTPResponse(status=200)
    else:
        return HTTPResponse(status=201)


@app.route('/install-certificates', method='POST')
def install_certificates():
    current_version = deploy_service_api.get_current_version(conduit_config.get_conduit_license_token())
    if not current_version:
        logger.warning("No current version in deployment database")
        return HTTPResponse(status=404, headers={'Content-Type': 'application/json'},
                            body={"error": "No version found"})
    user_session_json = dict()
    if request.headers['X-AUTH-USER']:
        user_session_json = json.loads(request.headers['X-AUTH-USER'])
    else:
        return HTTPResponse(status=403, headers={'Content-Type': 'application/json'},
                            body={"error": "No user found"})

    conduit_user_email = user_session_json['email']

    try:
        update_service = UpgradeService(deploy_service_api)
        # create update context with same version but with enabled_ssl and fqdn set
        ctx = update_service.create_upgrade_context(
            license_token=conduit_config.get_conduit_license_token(),
            current_version=current_version,
            dry_run=False,
            target_version=current_version,
            extra_vars={"enable_ssl": "true",
                        "conduit_fqdn": conduit_config.get_conduit_fqdn()}
        )

        update_service.install_certs_and_update(conduit_user_email=conduit_user_email,
                                                conduit_fqdn=conduit_config.get_conduit_fqdn(), ctx=ctx)
    except UpgradeError as ue:
        logger.exception("Failed to update ", exc_info=ue.cause)
        set_upgrade_failure_reason(ue.update_context, ue.cause)
        return HTTPResponse(status=500, body={"error": str(ue.cause)})
    except CertificateInstallationError as ce:
        logger.exception("Failed to install SSL certificates  ", exc_info=ce)
        return HTTPResponse(status=500, body={"error": str(ce)})
    except ConcurrentUpgradeAttempt:
        return HTTPResponse(status=409, headers={'Content-Type': 'application/json'},
                            body={"error": "upgrade is in progress"})
    except ValueError as ve:
        logger.exception("Illegal argument ", exc_info=ve)
        return HTTPResponse(status=400, headers={'Content-Type': 'application/json'},
                            body={"error": str(ve)})
    except BaseException as be:
        logger.exception("Failed to update ", exc_info=be)
        return HTTPResponse(status=500, headers={'Content-Type': 'application/json'},
                            body={"error": str(be)})

    return HTTPResponse(status=200)


@app.route('/management/services')
def get_services():
    response.headers['Content-Type'] = 'application/json'
    return json.dumps(__get_services_on_local_host())


@app.route('/management/cluster/services')
def get_cluster_services():

    host = {}
    host.update(__get_services_on_local_host())
    vms_hosts = conduit_config.get_deployment_vms().split()
    jwt = request.headers["Authorization"]
    for vm in vms_hosts:
        try:
            vm_services = __get_services_on_remote_vm(vm, jwt)
            host.update(vm_services)
        except Exception as ex:
            logger.warning("Could not retrieve data for vm {}.".format(vm), exc_info=ex)
            pass
    response.headers['Content-Type'] = 'application/json'
    return json.dumps(host)


@app.route('/management/services/<service_name>/status')
def get_service_status(service_name:str):
    try:
        service_status = _service_management.service_status(service_id=ServiceId(service_name))
        response.headers['Content-Type'] = 'application/json'
        return json.dumps(service_status.to_dict())
    except NoSuchService as not_found:
        return HTTPResponse(status=404)


@app.route('/management/services/<service_name>/restart', method='PUT')
def restart_service(service_name: str):

    try:
        service_id = ServiceId(service_name)
        service_status = _service_management.service_status(service_id=service_id)
        if service_status.is_restarting():
            return HTTPResponse(status=200)

        try:
            _service_management.stop_service(service_id=service_id)
        except RuntimeError as stop_error:
            logger.error("failed to stop service %s during restart procedure ".format(str(service_id)),
                         exc_info=stop_error)

        _service_management.start_service(service_id=service_id)
        return HTTPResponse(status=200)

    except NoSuchService as not_found:
        return HTTPResponse(status=404)
    except RuntimeError as re:
        logger.error("Error while ", exc_info=re)
        return HTTPResponse(status=500)


@app.route('/management/services/<service_name>/logs', method='GET')
def service_tail_log(service_name: str):
    follow_log = bool(request.query.follow)
    response.headers['Content-Type'] = 'text/plain'
    return _service_management.service_logs(service_id=ServiceId(name=service_name), follow=follow_log)


@app.route('/backup/repository/local/location', method='GET')
def get_local_backup_location():
    return HTTPResponse(status=200, body=json.dumps({ "location": conduit_config.get_system_snapshots_location()}))


@app.route('/backup/repository', method='POST')
def create_repository():
    body = request.json or dict()
    try:
        validate_response = _backup_service.validate_repository_request(body)
        if validate_response is not None:
            return HTTPResponse(status=400, body={"error": validate_response})
        _backup_service.create_repository_and_policy(body)
    except ConnectionTimeout as es:
        logger.warning("ConnectionTimeout {}".format(str(es)))
        return HTTPResponse(status=408, body={"error": "Connection timeout."})
    except ConnectionError as ce:
        logger.warning("Connection error. {}".format(str(ce)))
        return HTTPResponse(status=ce.status_code, body={"error": "Connection error"})
    except RepositoryVerificationError as rve:
        logger.warning("Repository verification error. {}".format(str(rve)))
        return HTTPResponse(status=rve.status_code, body={"error": "Invalid repository configuration. "
                                                                   "Please reconfigure your repository."})
    except ElasticsearchException as ee:
        logger.warning("ElasticsearchException. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Elasticsearch exception. "
                                                       "Failed to create repository. Please contact administrator for more info."})
    except errors.APIError as be:
        logger.warning("Docker error. {}".format(str(be)))
        return HTTPResponse(status=be.status_code, body={"error": be.explanation})

    return HTTPResponse(status=200)


@app.route('/backup/repository', method='GET')
def get_repository():
    try:
        es_repo = _backup_service.get_es_repository()
        if es_repo is None:
            return HTTPResponse(status=200, body=json.dumps([]))
        else:
            return HTTPResponse(status=200, body=json.dumps(es_repo))
    except ConnectionTimeout as es:
        logger.warning("ConnectionTimeout {}".format(str(es)))
        return HTTPResponse(status=408, body={"error": "Connection timeout."})
    except ConnectionError as ce:
        logger.warning("Connection error. {}".format(str(ce)))
        return HTTPResponse(status=ce.status_code, body={"error": "Connection error"})
    except RepositoryVerificationError as rve:
        logger.warning("Repository verification error. {}".format(str(rve)))
        return HTTPResponse(status=rve.status_code, body={"error": "Invalid repository configuration. "
                                                                   "Please reconfigure your repository."})
    except ElasticsearchException as ee:
        logger.warning("ElasticsearchException. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Elasticsearch exception. "
                                                       "Failed to retrieve repository. Please contact administrator for more info."})
    except Exception as ee:
        logger.warning("Repository exception. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Exception. Could not retrieve repository. "
                                                       "Please contact administrator for more info."})


@app.route('/backup/repository/<repo_name>', method='GET')
def get_repository(repo_name: str):
    try:
        es_repo = _backup_service.get_es_repository(repo_name)
        if es_repo is None:
            return HTTPResponse(status=200, body=json.dumps([]))
        else:
            return HTTPResponse(status=200, body=json.dumps(es_repo))
    except ConnectionTimeout as es:
        logger.warning("ConnectionTimeout {}".format(str(es)))
        return HTTPResponse(status=408, body={"error": "Connection timeout."})
    except ConnectionError as ce:
        logger.warning("Connection error. {}".format(str(ce)))
        return HTTPResponse(status=ce.status_code, body={"error": "Connection error"})
    except RepositoryVerificationError as rve:
        logger.warning("Repository verification error. {}".format(str(rve)))
        return HTTPResponse(status=rve.status_code, body={"error": "Invalid repository configuration. "
                                                                   "Please reconfigure your repository."})
    except ElasticsearchException as ee:
        logger.warning("ElasticsearchException. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Elasticsearch exception. "
                                                       "Failed to retrieve repository. Please contact administrator for more info."})
    except Exception as ee:
        logger.warning("Repository exception. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Exception. Could not retrieve repository. "
                                                       "Please contact administrator for more info."})


@app.route('/backup/snapshot/<repository_name>/<snapshot_name>', method='POST')
def create_snapshot(repository_name, snapshot_name):
    snapshot_settings = {
        "indices": ",".join(conduit_config.get_system_backup_indexes()),
        "ignore_unavailable": True,
        "wait_for_completion": True,
        "include_global_state": False
    }
    try:
        _backup_service.create_snapshot(repository_name, snapshot_name, snapshot_settings)
        return HTTPResponse(status=200)
    except ConnectionTimeout as es:
        logger.warning("ConnectionTimeout {}".format(str(es)))
        return HTTPResponse(status=es.status_code, body={"error": "Connection timeout."})
    except ConnectionError as ce:
        logger.warning("Connection error. {}".format(str(ce)))
        return HTTPResponse(status=ce.status_code, body={"error": "Connection error"})
    except RepositoryVerificationError as rve:
        logger.warning("Repository verification error. {}".format(str(rve)))
        return HTTPResponse(status=rve.status_code, body={"error": "Invalid repository configuration. "
                                                                   "Please reconfigure your repository."})
    except RepositoryConcurrencyError as rce:
        logger.warning("Repository verification error. {}".format(str(rce)))
        return HTTPResponse(status=rce.status_code, body={"error": "Could not read repository data because the contents of the repository "
                                                                   "do not match its expected state. This is likely the result of either "
                                                                   "concurrently modifying the contents of the repository by a process other than "
                                                                   "this cluster or an issue with the repository’s underlying storage. "
                                                                   "The repository has been disabled to prevent corrupting its contents. "
                                                                   "To re-enable it save backup configuration in order to recreate the repository"})
    except ElasticsearchException as ee:
        logger.warning("ElasticsearchException. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Elasticsearch exception. "
                                                       "Failed to retrieve repository. Please contact administrator for more info."})
    except Exception as ee:
        logger.warning("Repository exception. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Exception. Could not retrieve repository. "
                                                       "Please contact administrator for more info."})


@app.route('/backup/slm/<policy_name>', method='PUT')
def create_lifecycle_management(policy_name):
    snapshot_settings = request.json or dict()
    try:
        _backup_service.create_lifecycle(policy_name, snapshot_settings)
        return HTTPResponse(status=200)
    except ConnectionTimeout as es:
        logger.warning("ConnectionTimeout {}".format(str(es)))
        return HTTPResponse(status=es.status_code, body={"error": "Connection timeout."})
    except ConnectionError as ce:
        logger.warning("Connection error. {}".format(str(ce)))
        return HTTPResponse(status=ce.status_code, body={"error": "Connection error"})
    except ElasticsearchException as ee:
        logger.warning("ElasticsearchException. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Elasticsearch exception. "
                                                       "Failed to retrieve repository. Please contact administrator for more info."})
    except Exception as ee:
        logger.warning("Repository exception. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Exception. Could not retrieve repository. "
                                                       "Please contact administrator for more info."})


@app.route('/backup/slm', method='GET')
def get_slm():
    try:
        es_repo = _backup_service.get_lifecycle("conduit-snapshot")
        if es_repo is None:
            return HTTPResponse(status=200, body=json.dumps([]))
        else:
            return HTTPResponse(status=200, body=json.dumps(es_repo))
    except ConnectionTimeout as es:
        logger.warning("ConnectionTimeout {}".format(str(es)))
        return HTTPResponse(status=es.status_code, body={"error": "Connection timeout."})
    except ConnectionError as ce:
        logger.warning("Connection error. {}".format(str(ce)))
        return HTTPResponse(status=ce.status_code, body={"error": "Connection error"})
    except RepositoryVerificationError as rve:
        logger.warning("Repository verification error. {}".format(str(rve)))
        return HTTPResponse(status=rve.status_code, body={"error": "Invalid repository configuration. "})
    except RepositoryConcurrencyError as rce:
        logger.warning("Repository verification error. {}".format(str(rce)))
        return HTTPResponse(status=rce.status_code, body={"error": "Could not read repository data because the contents of the repository "
                                                                   "do not match its expected state. This is likely the result of either "
                                                                   "concurrently modifying the contents of the repository by a process other than "
                                                                   "this cluster or an issue with the repository’s underlying storage. "
                                                                   "The repository has been disabled to prevent corrupting its contents. "
                                                                   "To re-enable it save backup configuration in order to recreate the repository"})
    except ElasticsearchException as ee:
        logger.warning("ElasticsearchException. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Elasticsearch exception. "
                                                       "Failed to retrieve repository. Please contact administrator for more info."})
    except Exception as ee:
        logger.warning("Repository exception. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Exception. Could not retrieve repository. "
                                                       "Please contact administrator for more info."})


@app.route('/backup/slm/execute-lifecycle/<repository_name>/<policy_name>', method='POST')
def execute_lifecycle(repository_name, policy_name):
    try:
        es_repo = _backup_service.execute_lifecycle(policy_name, repository_name)
        if es_repo is None:
            return HTTPResponse(status=404,
                                body={"error": "Could not execute snapshot on demand. Snapshot policy {} does not exist"
                                      .format(policy_name)})
        else:
            return HTTPResponse(status=200, body={"snapshot_name": es_repo})
    except ConnectionTimeout as es:
        logger.warning("ConnectionTimeout {}".format(str(es)))
        return HTTPResponse(status=es.status_code, body={"error": "Connection timeout."})
    except ConnectionError as ce:
        logger.warning("Connection error. {}".format(str(ce)))
        return HTTPResponse(status=ce.status_code, body={"error": "Connection error"})
    except RepositoryVerificationError as rve:
        logger.warning("Repository verification error. {}".format(str(rve)))
        return HTTPResponse(status=rve.status_code, body={"error": "Invalid repository configuration. "
                                                                   "Please reconfigure your repository."})
    except RepositoryConcurrencyError as rce:
        logger.warning("Repository verification error. {}".format(str(rce)))
        return HTTPResponse(status=rce.status_code, body={"error": "Could not read repository data because the contents of the repository "
                                                                   "do not match its expected state. This is likely the result of either "
                                                                   "concurrently modifying the contents of the repository by a process other than "
                                                                   "this cluster or an issue with the repository’s underlying storage. "
                                                                   "The repository has been disabled to prevent corrupting its contents. "
                                                                   "To re-enable it save backup configuration in order to recreate the repository"})

    except ElasticsearchException as ee:
        logger.warning("ElasticsearchException. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Elasticsearch exception. "
                                                       "Failed to retrieve repository. Please contact administrator for more info."})
    except Exception as ee:
        logger.warning("Repository exception. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Exception. Could not retrieve repository. "
                                                       "Please contact administrator for more info."})


@app.route('/backup/snapshots/<snapshot_name>', method='GET')
def get_snapshot(snapshot_name):
    try:
        snapshot_response = _backup_service.get_snapshots(snapshot_name)
        return HTTPResponse(status=200, body=json.dumps(snapshot_response))
    except ConnectionTimeout as es:
        logger.warning("ConnectionTimeout {}".format(str(es)))
        return HTTPResponse(status=es.status_code, body={"error": "Connection timeout."})
    except ConnectionError as ce:
        logger.warning("Connection error. {}".format(str(ce)))
        return HTTPResponse(status=ce.status_code, body={"error": "Connection error"})
    except ElasticsearchException as ee:
        logger.warning("ElasticsearchException. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Elasticsearch exception. "
                                                       "Failed to retrieve repository. Please contact administrator for more info."})
    except Exception as ee:
        logger.warning("Repository exception. {}".format(str(ee)))
        return HTTPResponse(status=500, body={"error": "Exception. Could not retrieve repository. "
                                                       "Please contact administrator for more info."})


def __get_services_on_local_host():
    hostname = socket.gethostname()
    service_ids = list(_service_management.services())
    statuses = []
    for service_id in service_ids:
        try:
            service = _service_management.service_status(service_id=service_id)
            statuses.append(service.to_dict())
        except NoSuchService as ne:
            logger.warning("Can't find service " + str(service_id), exc_info=ne)
            pass
    return {hostname: statuses}


def __get_services_on_remote_vm(vm_hostname: str, jwt: str = None):
    try:
        url = "http://{}:{}/management/services".format(vm_hostname, server_port)
        svc_response = requests.get(url, headers={'Authorization': jwt})
        return svc_response.json()

    except Exception as ex:
        return HTTPResponse(status=ex.status_code, body="Could not retrieve services data for {}".format(vm_hostname))


def set_upgrade_failure_reason(upgrade_context: UpgradeContext, error: RuntimeError):
    try:
        logger.exception("Updating failed  deployment status for deploymentToken=%s",
                         conduit_config.get_conduit_license_token(), exc_info=error)

        deploy_service_api.set_deployment_status(upgrade_context, DeploymentStatus.UPGRADE_FAILED)
    except BaseException as be:
        logger.exception("Failed to update deployment status: %s " % str(be), exc_info=be)
        pass


def _to_bool(value, default: bool):
    if value:
        return str(value).lower() == "true"
    else:
        return default


if __name__ == "__main__":
    import __main__ as main
    from ansible.utils.display import Display

    display = Display(verbosity=1)
    setattr(main, 'display', display)

    app.run(host=server_ip, port=server_port, server='paste')
