import os
import shutil
from datetime import datetime

from azure.core.exceptions import ResourceNotFoundError, ClientAuthenticationError
from docker.errors import ImageNotFound
from elasticsearch import ElasticsearchException, Elasticsearch

import conduit_config

from artifact_repo import ArtifactFetcher
from certificate_manager import CertificateInstaller
from common import DeploymentStatus, UpgradeContext
from conduit_ansible_api import AnsibleRunner
from infrastructure_api import DeploymentApiClient

from log_config import logger


class UpgradeError(RuntimeError):

    def __init__(self, _cause: BaseException, upgrade_context: UpgradeContext):
        self.cause = _cause
        self.update_context = upgrade_context


class ConcurrentUpgradeAttempt(RuntimeError):
    pass


class UpgradeService:

    def __init__(self, deploy_service_api: DeploymentApiClient):
        self._deploy_service_api = deploy_service_api
        self._elasticsearch_client = Elasticsearch(hosts=conduit_config.get_elasticsearch_host(), timeout=60)

    @staticmethod
    def _close_system_indexes(es_client, system_indexes_list: list):
        for index_name in system_indexes_list:
            try:
                es_client.indices.close(index=index_name, ignore_unavailable=True)
            except ElasticsearchException as ee:
                logger.warning("Failed to close index %s during restore procedure, cause: %s ", index_name, str(ee))
                pass

    @staticmethod
    def _open_system_indexes(es_client, system_indexes_list: list):
        for index_name in system_indexes_list:
            try:
                es_client.indices.open(index=index_name, ignore_unavailable=True)
            except ElasticsearchException as ee:
                logger.warning("Failed to close index %s during restore procedure, cause: %s ", index_name, str(ee))
                pass

    def switch_to_maintenance_mode(self):
        '''
            Swich conduit service into maintenance mode so data can be changed but only served
            :rtype:
        '''

        self._close_system_indexes(self._elasticsearch_client, conduit_config.get_system_indexes())

    def switch_from_maintenance_mode(self):
        """
        Swich conduit service back to normal operation mode maintenance mode so data can be changed but only served
        :rtype:
        """
        self._open_system_indexes(self._elasticsearch_client, conduit_config.get_system_indexes())

    def rollback(self, current_version, ctx: UpgradeContext):
        """
            Handles a failed upgrade attempt and tries to upgrade to current_version using deployment ansible script

        :param ctx: failed upgrade context
        :param current_version: version of service before a failed up
        :param ctx:
        :return:
        """
        logger.info("Rolling back Conduit application  to %s version " % current_version)
        rollback_context = UpgradeContext(run_id=ctx.run_id()+"_rollback", license_token=ctx.license_token(),
                                          extra_vars=ctx.extra_vars(),
                                          current_version=current_version, target_version=current_version,
                                          deployment_client=self._deploy_service_api, dry_run=False)

        self.execute_update_with_context(upgrade_context=rollback_context)

        logger.info("Rolling back has been finished , setting deployment status back to STARTED ")
        ctx.update(DeploymentStatus.STARTED, rollback_context.current_version())

    def execute_update(self, license_token, current_version, target_version, dry_run=False):

        upgrade_context = self.create_upgrade_context(current_version, dry_run, license_token, target_version)
        self.execute_update_with_context(upgrade_context=upgrade_context)

    def execute_update_with_context(self, upgrade_context: UpgradeContext):

        assert upgrade_context
        assert upgrade_context.run_id()
        assert upgrade_context.target_version()
        assert upgrade_context.current_version()

        logger.info("Starting Upgrade procedure to %s version , run_id=%s" % (upgrade_context.target_version(), upgrade_context.run_id()))
        upgrade_context.set_status(DeploymentStatus.UPGRADE_STARTED)

        self.try_fetch_artifacts(upgrade_context)
        try:
            update_service_yml = conduit_config.get_ansible_script_location(
                upgrade_run_id=upgrade_context.run_id(),
                playbook_name="update-compose-singlebox-service.yml")

            upgrade_extra_vars = upgrade_context.extra_vars()
            upgrade_extra_vars.update({"bde_version_input": upgrade_context.target_version(),
                                       "started_by_service": "true",
                                       "conduit_deployment_auth_token": conduit_config.get_deployment_token(),
                                       "docker_user": upgrade_extra_vars.get("docker_user",
                                                                             conduit_config.get("system.docker.user")),
                                       "docker_password": upgrade_extra_vars.get("docker_password",
                                                                                 upgrade_context.docker_password())
                                       }
                                      )
            local_hosts_inventory = conduit_config.get_ansible_inventory(upgrade_context.run_id())
            ansible_runner = AnsibleRunner(_playbook_path=update_service_yml,
                                           extra_vars=upgrade_extra_vars, check_mode=upgrade_context.dry_run,
                                           inventory_file=local_hosts_inventory)

            result = ansible_runner.run()
            if result != 0:
                raise UpgradeError(_cause=RuntimeError("Update script exited with errors"),
                                   upgrade_context=upgrade_context)

            # update status and version in Conduit Central
            upgrade_context.update(DeploymentStatus.UPGRADE_FINISHED, upgrade_context.target_version())
            if not upgrade_context.dry_run:
                logger.info("Deployment has been successfully upgraded to version=%s " % upgrade_context.target_version())
            else:
                logger.info("Deployment upgrade to version=%s  has verified in dry run mode" % upgrade_context.target_version())

        except RuntimeError as re:
            logger.exception("Error while running upgrade", exc_info=re)
            if not upgrade_context.dry_run:
                try:
                    self.rollback(upgrade_context.current_version(), upgrade_context)
                    upgrade_context.update(DeploymentStatus.UPGRADE_FAILED_RECOVERED, upgrade_context.current_version())
                except BaseException as be:
                    logger.exception("Error while running rollback procedure  ", exc_info=be)
                    # XXX we are actually not sure about version at this moment
                    # lets keep previous version to let upgrade be executed again
                    upgrade_context.update(DeploymentStatus.UPGRADE_FAILED, upgrade_context.current_version())
                    raise UpgradeError(_cause=re, upgrade_context=upgrade_context)
            else:
                logger.info("Dry run upgrade failed , nothing to rollback: %s " % str(re))
                raise UpgradeError(_cause=re, upgrade_context=upgrade_context)

        finally:
            self.cleanup_working_dir(upgrade_context)

    def install_certs_and_update(self, conduit_user_email: str, conduit_fqdn: str, ctx: UpgradeContext):

        try:
            self.try_fetch_artifacts(upgrade_context=ctx, fetch_scripts=True, fetch_images=False)
            certificate_installer = CertificateInstaller(email=conduit_user_email,
                                                         conduit_fqdn=conduit_fqdn, ctx=ctx)

            certificate_installer.try_install()
            self.execute_update_with_context(ctx)
        finally:
            self.cleanup_working_dir(upgrade_context=ctx)

    @staticmethod
    def cleanup_working_dir(upgrade_context):
        try:
            ansible_dir = conduit_config.get_ansible_script_location(upgrade_run_id=upgrade_context.run_id())
            logger.info("Cleaning up working dir %s ", ansible_dir)
            shutil.rmtree(ansible_dir)

        except BaseException as be:
            logger.exception("Failed to cleanup working dir ", exc_info=be)
            pass

    def try_fetch_artifacts(self, upgrade_context, fetch_images=True, fetch_scripts=True):
        try:
            docker_user = upgrade_context.extra_vars().get("docker_user", conduit_config.get("system.docker.user"))
            docker_password = upgrade_context.docker_password()
            azure_account_key = upgrade_context.azure_storage_key()
            fetcher = ArtifactFetcher(azure_storage_key=azure_account_key, docker_password=docker_password,
                                      bp_docker_user=docker_user)
            self.fetch_artifacts(upgrade_context.target_version(), upgrade_context, fetcher=fetcher,
                                 fetch_images=fetch_images, fetch_scripts=fetch_scripts)
            assert os.path.exists(conduit_config.get_ansible_script_location(
                upgrade_run_id=upgrade_context.run_id())), "Can't find ansible script directory "

        except ResourceNotFoundError as ae:
            logger.exception("Failed to find conduit artifacts for version %s" % upgrade_context.target_version(), exc_info=ae)
            raise UpgradeError(_cause=ae, upgrade_context=upgrade_context)
        except ClientAuthenticationError as ce:
            logger.exception("Invalid Azure access key is provide to get conduit artifacts for version %s" % upgrade_context.target_version(), exc_info=ce)
            raise UpgradeError(_cause=ce, upgrade_context=upgrade_context)
        except ImageNotFound as inf:
            logger.exception("Failed to find docker iamges for version %s" % upgrade_context.target_version(), exc_info=inf)
            raise UpgradeError(_cause=inf, upgrade_context=upgrade_context)
        except AssertionError as assert_error:
            raise UpgradeError(_cause=assert_error, upgrade_context=upgrade_context)

    def create_upgrade_context(self, current_version, dry_run, license_token, target_version, extra_vars: dict = None):
        today = datetime.now()
        if dry_run:
            update_run_id = today.strftime('dry_run_conduit_update_%Y%m%d_%H-%M-%S').format()
        else:
            update_run_id = today.strftime('conduit_update_%Y%m%d_%H-%M-%S').format()
        upgrade_context = self._init_upgrade(license_token, update_run_id, current_version,
                                             target_version, dry_run, provided_env_vars=extra_vars)
        return upgrade_context

    def _init_upgrade(self, license_token: str, upgrade_run_id: str, current_version: str, target_version: str,
                      dry_run: bool, provided_env_vars: dict = None):
        logger.info("Started update procedure on %s, dry_run=%s" % (license_token, dry_run))
        current_status = self._deploy_service_api.get_deployment_status(license_token=license_token)
        if current_status and current_status.is_in_progress():
            raise ConcurrentUpgradeAttempt()

        version_response = self._deploy_service_api.get_version_def(target_version)
        extra_vars = version_response.get("extra_vars") or dict()
        if provided_env_vars:
            extra_vars.update(provided_env_vars)

        upgrade_ctx = UpgradeContext(run_id=upgrade_run_id, license_token=license_token, extra_vars=extra_vars,
                                     current_version=current_version, target_version=target_version,
                                     deployment_client=self._deploy_service_api, dry_run=dry_run)
        return upgrade_ctx

    @staticmethod
    def fetch_artifacts(target_version, upgrade_context: UpgradeContext, fetcher: ArtifactFetcher,
                        fetch_images=True, fetch_scripts=True):
        assert upgrade_context.target_version()

        if fetch_scripts:
            assert upgrade_context.azure_storage_key()
            archive_dest_path = conduit_config.get_ansible_archive_dest(upgrade_run_id=upgrade_context.run_id())
            ansible_scripts_dir = conduit_config.get_ansible_script_location(upgrade_run_id=upgrade_context.run_id())
            if os.path.exists(ansible_scripts_dir):
                shutil.rmtree(ansible_scripts_dir)

            os.makedirs(ansible_scripts_dir, exist_ok=True)
            fetcher.download_deployment_artifacts(target_version, archive_dest_path)
            logger.info("Unpacking Ansible scripts from %s ", archive_dest_path)
            shutil.unpack_archive(filename=archive_dest_path, extract_dir=ansible_scripts_dir)

        if fetch_images:
            assert upgrade_context.docker_password()
            fetcher.pull_docker_images(target_version=target_version)
